from numpy import *


def Pot(G,M,a,b,R,z):
   '''
   Miyamoto-Nagai Potential
   '''
   return -G*M/sqrt( R**2 + (a + sqrt(z**2+b**2))**2 )


def Vcirc(G,M,a,b,R):
   '''
   Miyamoto-Nagai Circular velocity
   '''
   return sqrt( G*M / ( R**2 + (a + b)**2 )**1.5 * R**2 )
   
def Omega(G,M,a,b,R):
   '''
   Miyamoto-Nagai Omega
   '''
   return sqrt( G*M / ( R**2 + (a + b)**2 )**1.5 )
   
def Kappa(G,M,a,b,R):
   '''
   Miyamoto-Nagai Kappa
   '''
   
   r2 = R*R
   x=sqrt(r2+(a+b)**2)
   return sqrt( G*M * (-3.*r2/x**5+4./x**3))
   

def Density(G,M,a,b,R,z):
   '''
   Miyamoto-Nagai Density
   '''
   
   
   zb   = sqrt(z**2+b**2)
   azb2 = (a+zb)**2
   R2   = R*R

   cte  = (b*b*M)/(4*pi) 
   return cte * (a*R2 + (a+3*zb)*azb2 ) / ( (R2+azb2)**2.5 * zb**3)
   

   
def SurfaceDensity(G,M,a,b,R):
   '''
   Miyamoto-Nagai Surface density
   '''
   
   z  = 0
   dz = b/100.
   integral = 0.0
   err = 1.
   i=0
   
   while (i<300):
     integral = integral+ Density(G,M,a,b,R,z)
     z=z+dz
     
     
     #err = abs(old_integral-integral)/integral
          
     i = i + 1
          
   return integral*2.*dz

   
   
